style = {'base_mpl_style': 'fast',
         'marketcolors': {'candle': {'up': '#082865', 'down': '#ae0019'},
                          'edge'  : {'up': '#082865', 'down': '#ae0019'},
                          'wick'  : {'up': '#082865', 'down': '#ae0019'},
                          'ohlc'  : {'up': '#082865', 'down': '#ae0019'},
                          'volume': {'up': '#082865', 'down': '#ae0019'},
                          'vcedge': {'up': '#1f77b4', 'down': '#1f77b4'},
                          'vcdopcod': False,
                          'alpha': 0.9},
         'mavcolors': None,
         'facecolor': None,
         'gridcolor': None,
         'gridstyle': None,
         'y_on_right': True,
         'rc': {'axes.edgecolor': '#082865',
                'axes.grid'     :  True,
                'axes.grid.axis': 'y',
                'grid.color'    : '#082865',
                'grid.linestyle': '--'
               },
         'base_mpf_style': 'starsandstripes'
        }
